function Particules = FastSlamUpdate(Particules, U, Z, MotionNoise, MeasurementNoise, dt)
    nParticules = size(Particules,2);
    DebugFig = false;
    if (DebugFig)
        figure(2)
        clf;
        hold on;
    end
    
    % On simule les dplacements
    for iParticule = 1:nParticules
        
        % Propagation avec bruit
        Unoisy(1) =  U(1) + MotionNoise.sV*randn();
        Unoisy(2) =  U(2) + MotionNoise.sOmega*randn();
        
        X = Particules(iParticule).X;
        X(1) = X(1) -Unoisy(1)/Unoisy(2)*sin(X(3)) + Unoisy(1)/Unoisy(2)*sin(X(3)+Unoisy(2)*dt);
        X(2) = X(2) +Unoisy(1)/Unoisy(2)*cos(X(3)) - Unoisy(1)/Unoisy(2)*cos(X(3)+Unoisy(2)*dt);
        X(3) = X(3) +Unoisy(2)*dt;
        
        % Que faire quand c'est la premire fois qu'on observe un landmark?
        for iZ = 1:size(Z,2)
            iLandmark = Z(3,iZ);
            muLandmark = Particules(iParticule).LandmarkMu{iLandmark};
            rangeZhat = norm(muLandmark-X(1:2));
            angleZhat = atan2((muLandmark(2)-X(2)),(muLandmark(1)-X(1)));
            
            Particules(iParticule).w = Particules(iParticule).w* ...
                normpdf(Z(1),rangeZhat,MeasurementNoise.sRange)* ...          
                normpdf(Z(2),angleZhat,MeasurementNoise.sAngle);          
        end
        
        Particules(iParticule).X = X;
        %Symbol = {'ro', 'bo', 'go'};
        %plotParticulesState(Particules(iParticule),iParticule,Symbol{iParticule});
      
    end
    
    % Normalize
    normConstant = 1/sum([Particules(:).w]);
    for iParticule = 1:nParticules
        Particules(iParticule).w = Particules(iParticule).w*normConstant;
    end
    
    % Resample
    indexCopy = ParticuleResampling([Particules(:).w],1)
    if (~isempty(indexCopy))
        % resampling is happening
        Particules = Particules(indexCopy);
        for iParticule=1:nParticules
            Particules(iParticule).w = 1/nParticules;
        end
    else
        % Degenerate case : on reboost les poids
        for iParticule=1:nParticules
            Particules(iParticule).w = 1/nParticules;
        end        
    end
    
    % Update of the landmarks individually, with EKF
    for iParticule = 1:nParticules
        %clf
        %plotParticulesState(Particules(iParticule),iParticule,'.')
        robotPose = Particules(iParticule).X;
        
        % Que faire quand c'est la premire fois qu'on observe un landmark?
        for iZ = 1:size(Z,2)
            iLandmark = Z(3,iZ);
            X = Particules(iParticule).LandmarkMu{iLandmark};
            P = Particules(iParticule).LandmarkSigma{iLandmark};
        
            rangeZhat = norm(X-robotPose(1:2));
            dx = X(1)-robotPose(1);
            dy = X(2)-robotPose(2);
            angleZhat = atan2(dy,dx)-robotPose(3);
            % https://www.cs.utexas.edu/~kuipers/slides/L17-FastSLAM.pdf
            z = Z(1:2,iZ);
            zhat = [rangeZhat angleZhat]';
            H = [ dx/rangeZhat  dy/rangeZhat; -dy/(rangeZhat^2) dx/(rangeZhat^2)];
            Cw = [MeasurementNoise.sRange^2 0; 0 MeasurementNoise.sAngle^2];
            
            % ======== Mise-a-jour ========
            K = P*H'/(H*P*H'+Cw);   
            r = (z-zhat);           
            r(2) = wrapToPi(r(2));
            X = X + K*r;
            P = (eye(2)-K*H)*P;
    
            Particules(iParticule).LandmarkMu{iLandmark} = X;
            Particules(iParticule).LandmarkSigma{iLandmark} =P;
        end
        %plotParticulesState(Particules(iParticule),iParticule,'x')
    end
end
