% Exemple de code pour FastSLAM
% (c) 2016 Maxime Latulippe et Philippe Giguere
% Cours GLO-4001/GLO-7021 Introduction a la robotique mobile
% Universite Laval, Quebec, Qc, Canada

clearvars
figure(1)

MakeAvi     = false;  % Pour creer un film avi. Normalement false.
DriveManual = true;   % Pour selectionner les waypoints manuellement
ShowCovariance = true; % Pour montrer la matrice de covariance
Nproche = 0; % Pour s'assurer de toujours avoir quelques points autour de la position de depart

% Ceci est fait pour qu'un instructeur puisse creer un film montrant
% l'evolution du systeme en fonction du temps
if (MakeAvi)
    aviobj = avifile('toto7.avi')
    aviobj.Fps = 5;
    aviobj.KeyFramePerSec = 5;
end

% parametres du systeme
max_range = 10;     % Portee du capteur
map_width = 20;    % Largeur de la carte
map_height = 20;   % Hauteur de la carte
N = 6;             % Nombre de points de reperes sur la carte
dt = 1;            % Intervalle de temps
Speed = 0.5;       % Vitesse lineaire du robot

% Initialisation du robot
robotPos = [0 0 0]';

% Positionnement aleatoire des reperes dans l'environnement
x = map_width*rand(1,N) - map_width/2;
y = map_height*rand(1,N) - map_height/2;
if (Nproche~=0)
    x(1:Nproche) = 4*rand(1,Nproche);
    y(1:Nproche) = 4*rand(1,Nproche)-2;
end
features = [x; y];

% bruits
sV = 0.02;
sW = 0.1*pi/180;
sRange = 2;
sAngle = 10.0*pi/180;

% Premier trace de la carte
clf
axis([-map_height/2 map_height/2 -map_height/2 map_height/2])
hold on
plot(features(1,:), features(2,:), '.k')
plot(robotPos(1), robotPos(2), 'bo')

IterationNum = 1;  % Iteration ici signifie waypoint

Trajectory = robotPos(1:2);

% Initialisation du filtre  particule
nParticules = 3;
for iParticule=1:nParticules
    Particules(iParticule).X = robotPos; %[randn() randn() wrapToPi(randn())]';
    Particules(iParticule).w = 1/nParticules;
    for iLandmark = 1:N
        if (0)
            Particules(iParticule).LandmarkMu{iLandmark} = features(:,iLandmark) + 3*randn(2,1);
            Particules(iParticule).LandmarkSigma{iLandmark} = eye(2)*40;
        else
            Particules(iParticule).LandmarkMu{iLandmark}(1,1) = map_width*rand() - map_width/2;
            Particules(iParticule).LandmarkMu{iLandmark}(2,1) = map_height*rand() - map_height/2;
            Particules(iParticule).LandmarkSigma{iLandmark} = eye(2)*100;
        end
        %Particules(iParticule).LandmarkMu{iLandmark} = NaN; % NaN indique landmark non-vu  date.
    end
end
MotionNoise.sOmega = sW*2;
MotionNoise.sV = sV*2;
MeasurementNoise.sAngle = 0.6;
MeasurementNoise.sRange = 2;


while (IterationNum<1000) 
    display(sprintf('Iteration %d',IterationNum));
    if (DriveManual)
        figure(1);
        display('Cliquez dans l image pour indiquer prochain waypoint');
        [Xtarget Ytarget] = ginput(1);
    else
        if (rand<0.07)
            % Once in a while go to home base, to reduce uncertainty
            Xtarget = 0;
            Yyarget = 0;
        else
            Xtarget = 20*rand-10;
            Ytarget = 20*rand-10;
        end
    end
    Trajectory(:,IterationNum+1) = [Xtarget; Ytarget];
    StepSize = 0.1;
    U = [];
    Straight = [];
    DesiredHeading = atan2(Ytarget-robotPos(2),Xtarget-robotPos(1));
    DeltaHeading = wrapToPi(DesiredHeading-robotPos(3));
    NRotStep = ceil(abs(DeltaHeading)/0.4);
    U(1,1:NRotStep) = 0.0;
    U(2,1:NRotStep) = DeltaHeading/NRotStep; % To make sure we have bang-on rotation;
    
    Distance = sqrt((Xtarget-robotPos(1))^2+(Ytarget-robotPos(2))^2);
    NStraight = ceil(Distance/Speed);
    Straight(1,1:NStraight) = Speed;
    Straight(2,1:NStraight) = 0.001; % Ne peut pas etre zero, sinon jacobienne kaput
    
    U = cat(2,U,Straight);

    for cmd = 1:size(U,2)
        
        % L'impression des graphiques se fait a des moments precis, pour
        % reduire le workload
        if  (DriveManual) || (IterationNum < 4) || ((cmd==1) && (IterationNum<40) ) || ( (cmd==1) && (mod(IterationNum,20) == 0) && (IterationNum>50) )
            figure(1)
            clf
            axis([-map_height/2 map_height/2 -map_height/2 map_height/2])
            hold on
            hl = line(Trajectory(1,:),Trajectory(2,:));
            set(hl,'Color',[0.7 0.7 0.7]);
            plot(features(1,:), features(2,:), '.k');
            for iFeature = 1:size(features,2)
                htext = text(features(1,iFeature)+0.3, features(2,iFeature)+0.3,sprintf('%d',iFeature));
                set(htext,'FontSize',16);
            end
            plot(robotPos(1), robotPos(2), 'ko');
            line([robotPos(1) robotPos(1)+0.5*cos(robotPos(3))], [robotPos(2) robotPos(2)+0.5*sin(robotPos(3))], 'color', 'k');
            plotCircle(robotPos(1), robotPos(2), max_range, 'k');
            
            for iParticule=1:nParticules
                plotParticulesState(Particules(iParticule),iParticule,'.');
            end

            title(sprintf('Trajectoire rectiligne %d',IterationNum),'FontSize',12);
            set(gca,'FontSize',12);
            axis([-15 15 -15 15]);
            drawnow();
            if (MakeAvi)
                % The figure must be in the first screen with dual monitor
                F = getframe(gcf);
                aviobj = addframe(aviobj,F);
            end
        end
        
        for iParticule = 1:size(Particules,2)
            figure(iParticule+2)
            clf
            plotParticulesState(Particules(iParticule),iParticule,'.');
            title(sprintf('Particule %d', iParticule));
            axis([-15 15 -15 15]);
        end

        if (cmd>80)
            display('cas probleme');
        end
        % deplacecment du robot
        v = U(1,cmd) + sV*randn;
        w = U(2,cmd) + sW*randn;
        robotPos(1) = robotPos(1) + -v/w*sin(robotPos(3)) + v/w*sin(robotPos(3)+w*dt);
        robotPos(2) = robotPos(2) + v/w*cos(robotPos(3)) - v/w*cos(robotPos(3)+w*dt);
        robotPos(3) = robotPos(3) + w*dt;

        % mesures prises par le robot
        relativePos = features - repmat(robotPos(1:2,:),1,N);
        ranges = sqrt(sum(relativePos.^2,1));
        % conserve uniquement les features visibles selon portee du scanner
        visible = find(ranges <= max_range);
        measures = relativePos(:,visible);
        ranges = ranges(visible);
        angles = atan2(measures(2,:), measures(1,:)) - robotPos(3);
        %display('angles avant bruit')
        %angles
        % ajout de bruit sur les mesures
        ranges = ranges + sRange*randn(1,numel(ranges));
        angles = angles + sAngle*randn(1,numel(angles));
        %display('angles APRES bruit')
        % angles

        % commande executee
        u = U(:,cmd);

        Z = [];
        Z(1,:) = ranges;
        Z(2,:) = angles;
        Z(3,:) = visible;
        
        Particules = FastSlamUpdate(Particules, u, Z, MotionNoise, MeasurementNoise, dt);
      
    end
    IterationNum = IterationNum + 1;
end


if (MakeAvi)
   aviobj = close(aviobj)
end